/****************************************************************
/*  Program Name: AX25.C
/*       Version: 1.1.0
/*          Date: 3 November 2001
/*
/*    Routines for sending AX.25 Data via MX614 modem.
/*
/*
/*	This software is available only for non-commercial  amateur radio
/*	or educational  applications.   ALL  other  uses  are  prohibited.
/*	This  software may  be  modified  only  if the  resulting code  be
/*	made available  publicly  and the  original  author given  credit.
/*	Contact the author for information on commercial licenses.
/*	(c) Copyright 1999,  John A. Hansen, W2FS.  All rights reserved.
/*
/*
/*	Many of the core functions were written for TAPR's PICe
/*	packet controller by John Hanson, W2FS (john@hansen.net).
/*	Some additions and modifications were made by Zack Clobes, 
/*	W0ZC (zack@clobes.net).
/*
/*
/***************************************************************/

//defines--------------------------------------------------------------
#define MAX_LEN_MESSAGE 90

//references to the output pin as used in the below functions
#byte PORT_C=7
#byte STATUS=3  // so it can access Status Register

#define BIT_DELAY 821	//This value may need to be adjusted to match any given oscilator.  821 seems to be pretty close for most resonators
#define TX_DELAY 45	//Number of 6.7ms delay cycles (i.e. TXDELAY)

//global variables-----------------------------------------------------
int i, stuff;
short flag, crcflag, slow;
byte crclo, crchi;

//function prototypes--------------------------------------------------
void writeIntDec(int iNumToSend, boolean bLeadingZero);
void writeLongDec(long lNumToSend, boolean bLeadingZero);
void sendSZString(char *szString);
void sendCrLf(void);
void flipOut(void);
void crcBit(byte tbyte);
void sendByte(byte inbyte);
void sendPacketStart(void);
void sendPacketEnd(void);
void initTNC(void);
void calibrate(void);
//---------------------------------------------------------------------
void flipOut(void) {			//flips the output state
	stuff = 0;
	// BE SURE TO MAKE THE BIT_TEST BIT MATCH THE OUTPUT BIT!!!!
	if (!bit_test(PORT_C, 2)) {
		output_high(OUT_DIG_TNC_DATA_TX);
	} else {
		output_low(OUT_DIG_TNC_DATA_TX);
	}
}
/////////////////////////////////////////////////////////////////////////////////
void crcBit(byte tbyte) {
#asm
	BCF		03,0
	RRF		crchi,F    //rotates the entire 16 bits to the right
	RRF		crclo,F
#endasm
	if (((STATUS & 0x01)^(tbyte)) == 0x01) {
		crchi = crchi^0x84;
		crclo = crclo^0x08;
	}
}
/////////////////////////////////////////////////////////////////////////////////
void sendByte(byte inbyte) {
	int k, bt;
	for (k=0;k<8;k++){
		bt = inbyte & 0x01;       //strip off the rightmost bit
		if ((crcflag == FALSE) && (flag == FALSE)) (crcbit(bt));   //do crc, but only if this is not a flag or crc byte
		if (!bt) (flipout());  // if this bit is a zero, flip the output state
		else {
			stuff++;    //increment 1's count
			if ((flag == FALSE) && (stuff == 5)) {   //stuff an extra 0, if 5 1's in a row
				delay_us(BIT_DELAY);
				flipout();
			}
		}
#asm
	RRF		inbyte,F
#endasm
	delay_us(BIT_DELAY);
	}
}
/////////////////////////////////////////////////////////////////////////////////
void sendPacketStart(void) {
//This function keys the transmitter, send the source and destination address, and gets ready
// to send the actual data.
	crclo=crchi=0xFF;
	stuff = 0;
	crcflag = FALSE;

	//init the MX614 for transmission
	output_low(OUT_DIG_TNC_DATA_TX);  //sets the packet output pin low
	output_high(OUT_DIG_TNC_M0);       //set up MX614 to transmit
	output_low(OUT_DIG_TNC_M1);
	delay_ms(100);        //wait for MX614 to wake up

	output_high(OUT_DIG_TNC_PTT);	//key the transmitter
	//tx delay
	flag = TRUE;
	for (i=0;i<TX_DELAY;i++) (sendByte(0x7E));		//adjust length for txdelay (each one takes 6.7ms)
	flag = FALSE;

	//begin transmission of actual packet
	//destination
	sendByte(0x82); //A
	sendByte(0xA0); //P
	sendByte(0xA4); //R
	sendByte(0xA6); //S 
	sendByte(0x40); //Sp
	sendByte(0x40); //Sp
	sendByte(0x60); //SSID=0

//**************  THIS IS WHERE THE CALLSIGN IS DETERMINED  **************
// The callsign is shifted to use the last seven bits of the byte - the first bit = 0
//  Use the table below to determine the hex values for your callsign.  A total of
//  six letters/numbers should be used.  For callsigns less than six digits, pad the
//  end of the callsign with spaces (0x40).
//
// The seventh byte is station ID (e.g. W0ZC-4's station ID is 4).  Again, use the table
//  below to determine the correct value for your station ID.
	//source
	sendByte(0xAE); //Byte 1 (W)
	sendByte(0x60); //Byte 2 (0)
	sendByte(0xAE); //Byte 3 (W)
	sendByte(0xA4); //Byte 4 (R)
	sendByte(0x40); //Byte 5 (Space)
	sendByte(0x40); //Byte 6 (Space)
	sendByte(0x77); //Station ID (0)

/*      Callsign byte lookup table
----------------------------------------------------------------
| Letters:                      | Numbers:                     |
| A = 0x82                      | 0 = 0x60                     |
| B = 0x84                      | 1 = 0x62                     |
| C = 0x86                      | 2 = 0x64                     |
| D = 0x88                      | 3 = 0x66                     |
| E = 0x8A                      | 4 = 0x68                     |
| F = 0x8C                      | 5 = 0x6A                     |
| G = 0x8E                      | 6 = 0x6C                     |
| H = 0x90                      | 7 = 0x6E                     |
| I = 0x92                      | 8 = 0x70                     |
| J = 0x94                      | 9 = 0x72                     |
| K = 0x96                      |                              |
| L = 0x98                      |------------------------------|
| M = 0x9A                      | Station ID:                  |
| N = 0x9C                      | 0 = 0x61                     |
| O = 0x9E                      | 1 = 0x63                     |
| P = 0xA0                      | 2 = 0x65                     |
| Q = 0xA2                      | 3 = 0x67                     |
| R = 0xA4                      | 4 = 0x69                     |
| S = 0xA6                      | 5 = 0x6B                     |
| T = 0xA8                      | 6 = 0x6D                     |
| U = 0xAA                      | 7 = 0x6F                     |
| V = 0xAC                      | 8 = 0x71                     |
| W = 0xAE                      | 9 = 0x73                     |
| X = 0xB0                      | 10 = 0x75                    |
| Y = 0xB2                      | 11 = 0x77                    |
| Z = 0xB4                      | 12 = 0x79                    |
|                               | 13 = 0x7B                    |
| Space = 0x40                  | 14 = 0x7D                    |
|                               | 15 = 0x7F                    |
----------------------------------------------------------------
End of lookup table
*/

	//overhead
	sendbyte(0x03);                          //send control byte -- 03 is APRS
	sendbyte(0xF0);                          // send PID
}
/////////////////////////////////////////////////////////////////////////////////
void sendPacketEnd(void) {
//This function closes out the packet with check-sums and unkeys the transmitter	
	//send the crc checksum	
	crcflag = TRUE;
	crclo =crclo^0xff;
	crchi = crchi^0xff;
	sendByte(crclo); //send the low byte of crc
	sendByte(crchi); //send the high byte of crc

	flag = TRUE;
	sendByte(0x7e);			  // Send a flag to end packet
	output_low(OUT_DIG_TNC_PTT);     //unkey PTT

	//set the status lines back to non-transmit mode
	output_low(OUT_DIG_TNC_DATA_TX); // set the transmit data line to low
	output_low(OUT_DIG_TNC_M0);   //put 614 back in receive mode
	output_high(OUT_DIG_TNC_M1);    
}
/////////////////////////////////////////////////////////////////////////////////
void initTNC(void) {
//This function initializes the modem chip.
// It should be called immediatly at powerup.
	output_high(OUT_DIG_TNC_M0);  //zero power mode
	output_high(OUT_DIG_TNC_M1);
	output_low(OUT_DIG_TNC_PTT);   //  PTT OFF
}
/////////////////////////////////////////////////////////////////////////////////
void writeIntDec(int iNumToSend, boolean bLeadingZero) {
//This function writes out an integer number to the packet
	int iCnt = 0;
	while (iNumToSend >= 100) {
		iNumToSend -= 100;
		iCnt++;
	}
	if (iCnt == 0) {
		if (bLeadingZero == TRUE) {
			sendByte('0');
		}
	} else {
		sendByte(iCnt + 48);
	}

	iCnt = 0;
	while (iNumToSend >= 10) {
		iNumToSend -= 10;
		iCnt++;
	}
	if (iCnt == 0) {
		if (bLeadingZero == TRUE) {
			sendByte('0');
		}
	} else {
		sendByte(iCnt + 48);
	}

	sendByte(iNumToSend + 48);
}
/////////////////////////////////////////////////////////////////////////////////
void writeLongDec(long lNumToSend, boolean bLeadingZero) {
//This function writes a long numeric data type to the packet
	int iCnt = 0;

	while (lNumToSend >= 10000) {
		lNumToSend -= 10000;
		iCnt++;
	}
	if (iCnt == 0) {
		if (bLeadingZero == TRUE) {
			sendByte('0');
		}
	} else {
		sendByte(iCnt + 48);
	}

	iCnt = 0;
	while (lNumToSend >= 1000) {
		lNumToSend -= 1000;
		iCnt++;
	}
	if (iCnt == 0) {
		if (bLeadingZero == TRUE) {
			sendByte('0');
		}
	} else {
		sendByte(iCnt + 48);
	}

	iCnt = 0;
	while (lNumToSend >= 100) {
		lNumToSend -= 100;
		iCnt++;
	}
	if (iCnt == 0) {
		if (bLeadingZero == TRUE) {
			sendByte('0');
		}
	} else {
		sendByte(iCnt + 48);
	}

	iCnt = 0;
	while (lNumToSend >= 10) {
		lNumToSend -= 10;
		iCnt++;
	}
	if (iCnt == 0) {
		if (bLeadingZero == TRUE) {
			sendByte('0');
		}
	} else {
		sendByte(iCnt + 48);
	}

	sendByte(lNumToSend + 48);  //always send the ones digit
}
/////////////////////////////////////////////////////////////////////////////////
void writeTime(long lTime) {
//This function uses a long int to write the time (hhmm) to the packet
	int iCnt = 0;

	iCnt = 0;
	while (lTime >= 1000) {
		lTime -= 1000;
		iCnt++;
	}
	sendByte(iCnt + 48);	//tens digit of hours

	iCnt = 0;
	while (lTime >= 100) {
		lTime -= 100;
		iCnt++;
	}
	sendByte(iCnt + 48);	//ones digit of hours

	iCnt = 0;
	while (lTime >= 10) {
		lTime -= 10;
		iCnt++;
	}
	sendByte(iCnt + 48);	//tens digit of minutes

	sendByte(lTime + 48);	//ones digit of minutes
}
/////////////////////////////////////////////////////////////////////////////////
void writeSignedIntDec(signed int iNumToSend, boolean bLeadingZero) {
//This function writes a signed integer to the packet
	int iCnt = 0;

	if (iNumToSend >= 0) {
		if(bLeadingZero == TRUE) {
			sendByte('+');
		}
	} else {
		sendByte('-');
		iNumToSend *= -1;
	}
	writeIntDec(iNumToSend, bLeadingZero);
}
/////////////////////////////////////////////////////////////////////////////////
void sendSZString(char *szString) {
//This function sends a null-terminated string to the packet
	while (*szString != 0) {
		sendByte(*szString);
		szString++;
	}
}
/////////////////////////////////////////////////////////////////////////////////
void sendCrLf(void) {
//This function sends a carriage return/line feed to the packet
	sendByte(10);
	sendByte(13);
}
/////////////////////////////////////////////////////////////////////////////////
void calibrate(void) {
	//goes into infinate loop to send 101010101010... on the output pin
	//  The period on the output should be 833uS or 1200Hz
	while (TRUE) sendByte(0);
}